/**
 **	undopainter.cpp
 **	UndoPainter
 **	by Saki Sakira <sakira@sun.dhis.portside.net>
 **	from 2006 June 7
 */

#include <stdio.h>

#include "undopainter.h"

UndoPainter::UndoPainter(int w, int h)
  : SketchPainter(w, h) {

  stroke_que = new pen_stroke[UndoStrokesMax];
  if (!stroke_que)
    printf("UndoPainter::stroke_que : mem. alloc. fails\n");

  temp_que = new pen_stroke[UndoStrokesMax];
  if (!temp_que)
    printf("UndoPainter::temp_que : mem. alloc. fails\n");

  inter_index = stroke_index = temp_index = 0;
}

UndoPainter::~UndoPainter(void) {
  delete[] stroke_que;
  delete[] temp_que;
}

void UndoPainter::changePainter(SketchPainter *sp) {
  int l = width * height >> 1;
  uint32* buf32 = (uint32*)frameBuffer();
  uint32* sbuf32 = (uint32*)(sp->frameBuffer());

  for (int i = 0; i < l; i ++)
    buf32[i] = sbuf32[i];

  inter_index = stroke_index = temp_index = 0;
}

void UndoPainter::enque(const pen_stroke &stroke) {
  fixStroke();

  if (temp_index < UndoStrokesMax) {
    temp_que[temp_index ++] = stroke;
  }
}

void UndoPainter::enque(const PtPen &pen, 
                        int x0, int y0, int x1, int y1) {
  pen_stroke ps;
  ps.pen = pen;
  ps.x0 = x0;
  ps.x1 = x1;
  ps.y0 = y0;
  ps.y1 = y1;
  enque(ps);
}

void UndoPainter::enquePenUp(void) {
  int l, diff, index;
  uint32 * stroke_array = (uint32 *)stroke_que;
  uint32 * temp_array = (uint32 *)temp_que;

  if (temp_index < UndoStrokesMax) {
    temp_que[temp_index ++].x0 = -1;
  }

  l = stroke_index + temp_index - UndoStrokesMax;
  for (index = 0; index < l; index ++)
    apply_stroke(stroke_que[index]);
    
  if (index) {
    while (stroke_que[index].x0 >=0 && index < (int)stroke_index)
      apply_stroke(stroke_que[index ++]);
    clearMask();
  }

  diff = (index * sizeof(pen_stroke)) >> 2;
  l = ((stroke_index - index) * sizeof(pen_stroke)) >> 2;
  for (int i = 0; i < l; i ++)
    stroke_array[i] = stroke_array[i + diff];

  diff = l;
  l = (temp_index * sizeof(pen_stroke)) >> 2;
  for (int i = 0; i < l; i ++)
    stroke_array[i + diff] = temp_array[i];

  stroke_index = stroke_index - index + temp_index;
  temp_index = 0;
  inter_index = stroke_index;

}

void UndoPainter::apply_stroke(pen_stroke & stroke) {
  if (stroke.x0 >= 0) {
    pen = stroke.pen;
    //printf("%d ", pen.brush_method);
    //    if (pen.brush_method == WaterBrush)
    //      pen.brush_method = SolidBrush;
    drawLine(stroke.x0, stroke.y0, stroke.x1, stroke.y1);
  } else {
    clearMask();
  }
}

void UndoPainter::getBuffer(uint16* dbuf, uint ratio) {
  /*  0 < ratio < 0x10000 */

  uint16* last_buf;

  uint32* dbuf32 = (uint32*)dbuf;
  uint32* buf32 = (uint32*)buf;

  uint index = (ratio * stroke_index) >> 16;

  int l = width * height >> 1;
  for (int i = 0; i < l; i ++)
    dbuf32[i] = buf32[i];

  last_buf = buf;
  buf = dbuf;
  for (uint i = 0; i < index; i ++)
    apply_stroke(stroke_que[i]);
  clearMask();
  buf = last_buf;
  
  inter_index = index;
}

void UndoPainter::fixStroke(void) {
  if (inter_index < stroke_index) {
    stroke_index = inter_index;
    enquePenUp();
  }
}

